#include "util.h"
#include "temps.h"
#include "ueib.h"
#include <jwsmtp/jwsmtp.h>
#include <fcntl.h>
#include <iomanip>
#include <fstream>

//BD Connection
#define HOST "localhost"
#define BASE "automate"
#define LOGIN "root"
#define PASS ""


// K8055 ana input
#define INP_TEMP 0 // temperature captor

// used for getting the k8055 config
config cfg; 

// used to give the more relevant temperature according to 10 values
// not the best perf but tab 10
double give_temperature(double t[10])
{
int i,j,n,max;
double v,r;

max=0;
for(i=0;i<10;i++)
{
	v=t[i];n=0;
	for(j=0;j<10;j++)
		{if (t[j]==v) n++;}
	if (n>=max) {max=n;r=v;}
}
return r;
}	

// used for restarting automaton
void restart_automaton(short getstart, connexion &c)
{
int n;

if (getstart==1)
	{
	// normal restart
	cout << "restarting automaton" << endl;
	c.insert_log("automaton restarted");
	n=fork();
	if (n==0) {execl("./automate","automate",NULL);}
	
}

if (getstart==2)
	{
	// restart with reinit of bd
	cout << "restarting automaton with fresh bd" << endl;
	c.insert_log("automaton restarted with reinit of bd");
	n=fork();
	if (n==0) {execl("./automate","automate","init-bd",NULL);}
	}

if (getstart==3)
	{
        // stopping
	cout << "stopping automaton" << endl;
	c.insert_log("automaton stopped");
	}

}

// send mail alert
void send_mail(string message, connexion &c)
{
time_t rawtime;
time(&rawtime);
int n;
string m;

cout << ctime (&rawtime) << "sending an email alert:"  << message << endl;
m = "ALERT: "+message;
c.insert_log(m);
n=fork();
if(n==0)
{
jwsmtp::mailer m(cfg.EXP.c_str(), cfg.DEST.c_str(), "ALERT !!!", message.c_str(), cfg.SMTP.c_str(),jwsmtp::mailer::SMTP_PORT, false);
   	//m.username("");
	//m.password("");
	m.send( ); // send the mail
	//m.response( ); // receiving error from the smtp server	
	exit(0);
}
}	

// insert error type in table error + log
void insert_error(connexion &c,short type)
{
c.insert_error(type);
c.insert_log("error: BD or K8055 Card");
}


//change output state - write log
void change_state(connexion &c,ueib_handle *uh, short out,short state,string s)
{
string st,st2,s2;
time_t rawtime;


time ( &rawtime );
s2=ctime(&rawtime);

st=st+ s2 + s ;
if (state==0) st2=" : off\n"; else st2=" : on\n";
st =st +st2;
cout << st<< endl; 
ueib_set_dig_output(uh,out, state);  
c.insert_log(s+st2);
c.maj_out(out,state);
}


//print status
void print_status(ueib_handle *uh)
{
double temperature;

cout << "**** status ****\n";
cout << "=>inputs:\n";
if (ueib_get_dig_input(uh, cfg.STOP_PUMP)==1) cout  << setw(17) << "button 1:stopping pump for 10 min\n";
if (ueib_get_dig_input(uh, cfg.STOP_SKIMMER)==1) cout << setw(17) << "stopping skimmer\n";
if (ueib_get_dig_input(uh, cfg.OSMOLATION)==1) cout << setw(17) << "osmolation activated by captor\n";
if (ueib_get_dig_input(uh, cfg.RESERV)==1) cout << setw(17) << "RESERV EMPTY !!!\n";
temperature =(double) (101*ueib_get_ana_input(uh,INP_TEMP)/256)-23;
cout << "temperature : " << temperature << endl;	

cout << "\n=>outputs:\n";
if (ueib_get_dig_output(uh, cfg.OUTHQI)==1) cout << setw(17) << "HQI working\n"; else cout <<"HQI stopped\n";
if (ueib_get_dig_output(uh, cfg.OUTBLUE)==1) cout << setw(17) <<"Blue light working\n"; else cout <<"Blue light stopped\n";
if (ueib_get_dig_output(uh, cfg.OUTPUMP)==1) cout << setw(17) <<"Pump working\n"; else cout <<"Pump stopped\n";
if (ueib_get_dig_output(uh, cfg.OUTSKIMMER)==1) cout << setw(17) << "Skimmer working\n"; else cout <<"Skimmer stopped\n";
if (ueib_get_dig_output(uh, cfg.OUTOSMOLATION)==1) cout << setw(17) << "osmolation working\n"; else cout <<"Osmolation stopped\n";
if (ueib_get_dig_output(uh, cfg.OUTRAH)==1) cout << setw(17) << "RAH pump working\n"; else cout <<"RAH pump stopped\n";
if (ueib_get_dig_output(uh, cfg.OUTTEMP)==1) cout << setw(17) << "Chauffage working\n"; else cout <<"Chauffage stopped\n";
if (ueib_get_dig_output(uh, cfg.OUTPERI)==1) cout << setw(17) << "Peristatic pump working\n"; else cout <<"Peristatic pump stopped\n";
}



 

//return the current state for one output
// depends of a param obj
short get_state(string s,connexion  &c)
{
temps t,tbegin,tend;
bool button,b;
short r;
param p;

c.get_parameter(s,p,0);

tbegin.set_time(p.get_init_timer());
tend.set_time(p.get_end_timer());
button=p.get_button();

t.set_current_time();
b=t.inclus(tbegin,tend);

// the different possible cases from the button and the timer
if(b==true && button==true) r=1;
if(b==true && button==false) r=0;
if(b==false) r=0;
return r;
}

//return the current state of the rah in database : hour X hour Y : enabling rah pump X min every Y min
// depends of a param obj
short get_state_rah(time_t &timer_rah,connexion  &c)
{
time_t t;
temps tY;
bool button,b=false;
short r;
param p;

c.get_parameter("rah",p,0);
tY.set_time(p.get_end_timer());
button=p.get_button();
time(&t);
if ( difftime(t,timer_rah) > (tY.get_hour()*3600+tY.get_min()*60)) b=true; 


// the different possible cases from the button and the timer
if(b==true && button==true) {r=1; time(&timer_rah);}
if(b==true && button==false) r=0;
if(b==false) r=0;
return r;
}

// getting state of peristatic pump (depends of 5 time elapses)
short get_state_peri(connexion  &c)
{
temps t,tbegin,tend;
bool button,b;
bool r[5];
short i;
param p;

t.set_current_time();
for(i=0;i<5;i++)
	{
	c.get_parameter("pdoseuse",p,i+1);
	tbegin.set_time(p.get_init_timer());
	tend.set_time(p.get_end_timer());
	button=p.get_button();
	b=t.inclus(tbegin,tend);
	if(b==true && button==true) r[i]=true;
	if(b==true && button==false) r[i]=false;
	if(b==false) r[i]=false;
	}

if ((r[0]||r[1]||r[2]||r[3]||r[4])==true) return 1; else return 0;
}



// enabling the rah pump X min
bool enable_rahpump(time_t &timer_rah,connexion  &c)
{
time_t t;
temps tX;
bool b=true;
short r;
param p;

c.get_parameter("rah",p,0);
tX.set_time(p.get_init_timer());
time(&t);
if ( difftime(t,timer_rah) > tX.get_min()*60) b=false; 

return b;
}



//*********************************************************
int main(int argc, char** argv)
{
int device = 0;
int parsed_options = 0;
connexion c(HOST,BASE,LOGIN,PASS);
// the state of output
short state_hqi=0;
bool stop_hqi=false; // hqi stop if temperature > THQI (user defined)
short state_blue=0;
short state_pump=0;
short state_skimmer=0;
short state_skimmer_t=1; //used for saving the skimmer state if water is too high 
short state_peri=0; // state of peristatic pump
short state_temp=0; // state of the ***chauffage***
short state_osmolation=0;
short state_high=0; // state of the level water
short captor_high=0; // state of the high captor
short stop_pump=0; // state of the physical button (input 1) for stopping pumps 
short stop_skimmer=0; // state of the physical button (input 2) for stopping pump skimmer
short init_osmolation=0; // state of osmolation
short empty_reserv=0; // state of the reserv water 1=empty => no osmolation !
short ret; // return error of the K8055
short state_rah=0; // state of the rah pump; enabled x min every M min
short getstart=0; // used for checking if the automaton must be restarted

short st,sh,r,day;
time_t rawtime, t_timer,t_timer2, t_sk, timer_temp,timer_rah, timer_hqi;

bool started=true; // used to start hqi 15 min after the automaton start
bool stop_p=false; // used for stopping pump
bool stop_sk=false; // used for stopping pump skimmer 2min
bool stop_osmolation=false; // used for stopping osmol
bool stop_rah=false; // used for stopping the rah pump
bool reserv_mail=false;
bool full=false;  // used to check if decant is full
bool new_m=false;// used to check new hour
bool sup30=false;
bool inf24=false;
bool sk_forced=false; // used to check if skimmer has been forced to run it if necessary
double dif,t2;
double temperature=25;


double te[10]; 
int count=0; // used to compute the temperature from 10 values retains the best one !!!
string s,s2,m;
char ch=0;


//******************************************************************
if(argc > 1)
	{	
	        if(!strcmp(argv[1],"init-bd") )
			{
			c.construct();
			insert_error(c,2);
			}
		

	}

fcntl(0,F_SETFL,O_NONBLOCK); //deblocking stdin

// database connection
c.connect_to_db();
sleep(2); // used to wait the end of another automaton process


// get the config
c.get_config(cfg);
c.init_out();





//init k8055 interface 0
device=0;
usb_init();
	ueib_handle *uh = ueib_init(device);
	if(uh == NULL)
	{
		printf("Error connecting to device\n");
		// error type 1 in table error
		insert_error(c,1);c.set_stop();c.close();exit(EXIT_FAILURE);
	}
	r=ueib_update(uh);
	if(r==-1) {insert_error(c,1);c.set_stop();c.close();exit(EXIT_FAILURE);} // can't read k8055 stopping


// init start to 0 in database models and indicates the starting of the automaton
c.set_start();


try {


cout << "*** running automate *** \n" << "*** p : print status ***\n"  << "*** q : quit ! ***\n";
send_mail("automate started",c);

//***************************************************************************
//init some ouputs
// for updating k8055
state_skimmer=1;
change_state(c,uh, cfg.OUTSKIMMER,state_skimmer,"skimmer changing");
ueib_update(uh);
//***************************************************************************

// init timer of the rah pump
time ( &timer_rah);
// init timer of the hqi
time ( &timer_hqi);

while(ch!= 'q' && getstart==0)
	{
	if(ch=='p') print_status(uh);
	
	//*******************************************************************
	// checking if the automaton must be restarted 
	getstart=c.get_start();
	restart_automaton(getstart,c);
	
	
	//*******************************************************************
	//getting input states
	stop_pump=ueib_get_dig_input(uh,cfg.STOP_PUMP);
	stop_skimmer=ueib_get_dig_input(uh,cfg.STOP_SKIMMER);
	init_osmolation=ueib_get_dig_input(uh,cfg.OSMOLATION);
	empty_reserv=ueib_get_dig_input(uh,cfg.RESERV);
	captor_high=ueib_get_dig_input(uh,cfg.HIGH);
	
	//******************************************************************* 
	// getting analogic input 1 => temperature 
	t2= (double)ueib_get_ana_input(uh,INP_TEMP);
	te[count] =(101*t2/256)-23;
	//cout << te[count] << "     " << endl;
	count++;
	if (count==10) 
	{
		temperature=give_temperature(te);
		count=0;
	}
	//cout << "     " << temperature << endl;
	
	//*******************************************************************
	// get the forced status of output 0:normal 1:stop forced 2:run forced 
	c.get_forced(cfg);
	//*******************************************************************
	// state of hqi changing ?
	if(cfg.forced_hqi!=0) 
		{
		if(state_hqi != cfg.forced_hqi-1)
			{
			state_hqi=cfg.forced_hqi-1;
			change_state(c,uh, cfg.OUTHQI,state_hqi,"hqi changing forced");
			}
		}
	else
	{
	if (temperature > cfg.THQI && stop_hqi==false && state_hqi==1)   // HQI stopped if t > THQI
		{
		stop_hqi=true;
		state_hqi=0;
		// state changing
		change_state(c,uh, cfg.OUTHQI,state_hqi,"ALERT: temperature too hot: hqi changing");
		}
	// get hqi parameters
	st=get_state("hqi",c);
	if (state_hqi != st && started== false && stop_hqi==false)
		{
		if (temperature > cfg.THQI)
			{
			stop_hqi=true;
			c.insert_log("ALERT: temperature too hot: hqi not started");
			}
		else
			{

			// state changing
			state_hqi=st;
			change_state(c,uh, cfg.OUTHQI,state_hqi,"hqi changing");
			}
		}
	// wait starthqi (user defined) before starting hqi 
	if (started == true)
		{
		time ( &rawtime );
		dif = difftime (rawtime,timer_hqi);
		if( (dif/60) > cfg.STARTHQI) 
			{
			started=false;
			}
		
		}
	}
	//*******************************************************************
	// state of blue changing ?
	if(cfg.forced_blue!=0)
		{
		 if(cfg.forced_blue-1 != state_blue)
			{
			state_blue=cfg.forced_blue-1;
			change_state(c,uh, cfg.OUTBLUE,state_blue,"blue light changing forced");
			}
		}
	else
	{
	st=get_state("blue",c);
	if (state_blue != st )
		{
		// state changing
		state_blue=st;
		change_state(c,uh, cfg.OUTBLUE,state_blue,"blue light changing");
		}
	//********************************************************************
	//state of pumps 
	//depends of param obj and of input 1 ( physical button)
	if(cfg.forced_pump!=0)
		{
		 if (cfg.forced_pump-1 !=state_pump)
			{
			state_pump=cfg.forced_pump-1;
			change_state(c,uh, cfg.OUTPUMP,state_pump,"pump changing forced");
			}
		}
	else
	{
	if (stop_pump==1)
		{
		//pumps must be stopped during 10 min 
		time (&t_timer);
		stop_p=true;
		}

	if(stop_p==true)
		{
		time ( &rawtime );
		dif = difftime (rawtime,t_timer);
		if( (dif/60) > 10.0) 
			{
			stop_p=false;
			}
		else
			{
			//stopping
			if (state_pump==1)
				{
				// state changing
				state_pump=0;
				change_state(c,uh, cfg.OUTPUMP,state_pump,"pump changing for 10min:");
				}
		
			}
		}	
	else
	{
	st=get_state("pump",c);
	if (state_pump != st )
		{
		// state changing
		state_pump=st;
		change_state(c,uh, cfg.OUTPUMP,state_pump,"pump changing");
		}
	}
	}
	}
	//********************************************************************************
	//enabling / stopping skimmer pump
	//stopping skimmer if the high captor is enabled
	if(cfg.forced_skimmer!=0) 
		{
		sk_forced=true;
		if (cfg.forced_skimmer-1 != state_skimmer)
			{
			state_skimmer=cfg.forced_skimmer-1;
			change_state(c,uh, cfg.OUTSKIMMER,state_skimmer,"skimmer changing forced");
			}
		}
	else
	{
	if(cfg.forced_skimmer==0 && sk_forced==true && state_skimmer==0 && stop_skimmer==0) 
		{
		state_skimmer=1;
		sk_forced=false;
		change_state(c,uh, cfg.OUTSKIMMER,state_skimmer,"skimmer changing");
		}
	
	if (captor_high==1 && state_high==0)
		{
		state_high=1;
		state_skimmer_t=state_skimmer;
		state_skimmer=0;
		change_state(c,uh, cfg.OUTSKIMMER,state_skimmer,"skimmer stopped: water level on the high captor");
		send_mail("water level on the high captor",c);
		}
	if (captor_high==0 && state_high==1)
		{
		state_high=0;
		state_skimmer=state_skimmer_t;
		change_state(c,uh, cfg.OUTSKIMMER,state_skimmer,"water level normal: skimmer returns on its previous state");
		}
	time(&rawtime);
	if(stop_sk==true && (difftime(rawtime,t_sk)>30)) {stop_sk=false;}
	if(stop_skimmer==1 && stop_sk==false) 
		{
		stop_sk=true;
		time(& t_sk);
		if (state_skimmer==1) state_skimmer=0; else state_skimmer=1;
		stop_skimmer=0;
		// state changing
		change_state(c,uh, cfg.OUTSKIMMER,state_skimmer,"skimmer pump changing");
		}
	}
	//*********************************************************************************
	//state of osmolations 
	if(cfg.forced_osmolation!=0)
		{
		if (cfg.forced_osmolation-1 != state_osmolation)
			{
			state_osmolation=cfg.forced_osmolation-1;
			change_state(c,uh, cfg.OUTOSMOLATION,state_osmolation,"osmolation changing forced");
			}
		}
	else
	{
	if (empty_reserv==0 && init_osmolation==1)
		{
		//osmolation pump 
		time (&t_timer2);
		stop_osmolation=true;
		reserv_mail=false;
		}

	if(stop_osmolation==true)
		{
		time ( &rawtime );
		dif = difftime (rawtime,t_timer2);
		if( (dif) >= cfg.TOSMOLATION) 
			{
			stop_osmolation=false;
			// stopping osmolation
			state_osmolation=0;
			change_state(c,uh, cfg.OUTOSMOLATION,state_osmolation,"osmolation stopped");
			//inserting time and elapse of osmolation in bd (osmolation)
			time ( &rawtime );
			s=ctime(&t_timer2);
			s2=ctime(&rawtime);
			c.insert_osmolation(s.substr(s.find(":",0)-2,8),s2.substr(s2.find(":",0)-2,8));

			}
		else
			{
			if(state_osmolation==0)
				{
				// state changing init osmolation if not running
				state_osmolation=1;
				change_state(c,uh, cfg.OUTOSMOLATION,state_osmolation,"osmolation ");
				}
		
			}
		}	
	
	if (empty_reserv==1 && reserv_mail==false)
	{
	// no waterin the reserv => sending an email alert
	send_mail("no water in the reserv !!!",c);	
	reserv_mail=true;
	if(state_osmolation==1)
		{
		stop_osmolation=false;
		// stopping osmolation
		state_osmolation=0;
		change_state(c,uh, cfg.OUTOSMOLATION,state_osmolation,"no water stopping osmolation");
		//inserting time and elapse of osmolation in bd (osmolation)
		time ( &rawtime );
		s=ctime(&t_timer2);
		s2=ctime(&rawtime);
		c.insert_osmolation(s.substr(s.find(":",0)-2,8),s2.substr(s2.find(":",0)-2,8));
		}

	}
	if (captor_high==1 && full==false)
	{
	// decantation is full, osmolation captor is not working => sending an email alert
	send_mail("decantation full of water, pb with osmolation captor !!!",c);	
	full=true;
	if(state_osmolation==1)
		{
		stop_osmolation=false;
		// stopping osmolation
		state_osmolation=0;
		change_state(c,uh, cfg.OUTOSMOLATION,state_osmolation,"decantation full of water, stopping osmolation");
		//inserting time and elapse of osmolation in bd (osmolation)
		time ( &rawtime );
		s=ctime(&t_timer2);
		s2=ctime(&rawtime);
		c.insert_osmolation(s.substr(s.find(":",0)-2,8),s2.substr(s2.find(":",0)-2,8));
		}

	}
       if (captor_high==0 && full==true)
	{
	// decantation is not  full, osmolation working
	full=false;
	}
	}
	//********************************************************************************
	//temperature state
	// maj current temperature
	c.insert_current_temperature(temperature);
	
	if (temperature > 30.0 && sup30==false)
		{
		send_mail("temperature > 30 !!!!!",c);
		sup30=true;
		}
	if (temperature <= 30 && sup30==true) sup30=false;
	
	if((cfg.forced_temp!=0) && (cfg.forced_temp-1 !=state_temp))
		{
		state_temp=cfg.forced_temp-1;
		change_state(c,uh, cfg.OUTTEMP,state_temp,"heating changing forced");
		}
	else
	{
	if (temperature > 28.0 && state_temp==1) 
		{
		state_temp=0;
		change_state(c,uh, cfg.OUTTEMP,state_temp,"stopping heating");
		}
	if (temperature < 26.0 && state_temp==0) 
		{
		state_temp=1;
		change_state(c,uh, cfg.OUTTEMP,state_temp,"enabling heating");		
		}
	}
	if (temperature < 23.0 && state_temp==1 && inf24==false)
		{
		send_mail("temperature < 23 !!! pb with chauffage",c);
		inf24=true;
		}
	if (temperature >= 24) inf24=false;

	//********************************************************************************
	// some operations depending on time
	time (&timer_temp);
	s=ctime(&timer_temp); 
	s2=s.substr(s.find(":",0)+1,2);
	st=atoi(s2.c_str()); // minute
	s2=s.substr(s.find(":",0)-2,2); 
	sh=atoi(s2.c_str()); //hour
	
	if (st==0 && new_m==false) 
		{
		new_m=true;
		if (sh==0)
			{
			//new day !!!
			stop_hqi=false; // re-enable HQI 
			c.maj_log(); // delete logs of the previous month
		
			m=s.substr(8,2);
			day=atoi(m.c_str());
			if(day==1) c.maj_temperature();  // new month => min and max temp deleted in temperature table
			}
	
		c.insert_temperature(temperature);
		c.save(); // save current bd every hour
		}
	if (st==1 && new_m==true)
		new_m=false;
		
        


		
	//********************************************************************************
	// rah pump state
	if(cfg.forced_rah!=0) 
		{
		if (cfg.forced_rah-1!=state_rah)
			{
			state_rah=cfg.forced_rah-1;
			change_state(c,uh, cfg.OUTRAH,state_rah,"rah changing forced");
			}
		}
	else
	{
	if (state_rah==0) {state_rah=get_state_rah(timer_rah,c);}
	
	if (state_rah==1 && stop_rah==false)
		{
		change_state(c,uh, cfg.OUTRAH,state_rah,"enabling rah pump");
		stop_rah=true;
		}
	if (stop_rah==true)
		{
		if (enable_rahpump(timer_rah,c)==false)
			{
			state_rah=0;
			change_state(c,uh, cfg.OUTRAH,state_rah,"stopping rah pump");
			stop_rah=false;
			time(&timer_rah);
			}
		
		}
	}	
	//********************************************************************************
	// state of peristatic pump
	if(cfg.forced_peri!=0) 
		{
		if (cfg.forced_peri-1!=state_peri)	
			{
			state_peri=cfg.forced_peri-1;
			change_state(c,uh, cfg.OUTPERI,state_peri,"peristatic pump changing forced");
			}
		}
	else
	{
	st=get_state_peri(c);
	if (state_peri != st )
		{
		// state changing
		state_peri=st;
		change_state(c,uh, cfg.OUTPERI,state_peri,"peristatic pump changing");
		}
	}
	//********************************************************************************
	
	r=ueib_update(uh);
	sleep(1);
	if(r==-1) // can't read k8055
	{
	 //error type 2 in table error
	insert_error(c,1);
	c.set_stop();c.close();
	exit(EXIT_FAILURE);
	}
	ch=0;
	read(0, &ch, 1);	
	}//endwhile
	
	c.set_stop();
	c.close();
	ueib_close(uh);



}
catch (mysqlpp::BadQuery& er) {
		// handle any connection or query errors that may come up
		cout << "Error: mysql";
		return -1;
	}
	

	return 0;
}


